/* =========================================================================== 
*  Crawling CPG
*
*  Copyright (c) 2005 Ludovic Righetti <ludovic.righetti _at_ a3.epfl.ch>
*
*  May be copied or modified under the terms of the GNU General Public
*  License.  See COPYING for more information.
*
*  Contributors:
*  @CONTNAME@ <@CONTEMAIL@> 
*
*
*
*  First implemented: 20/1/06 Ludovic Righetti 
* 
*  
=============================================================================*/
#include "Crawling_CPG.h"

#include <gsl/gsl_sf.h>
#include <gsl/gsl_math.h>

int Crawling_CPG::trot_matrix[4*4] = {0,-1,-1,1,
				       -1,0,1,-1,
				       -1,1,0,-1,
				       1,-1,-1,0};

int Crawling_CPG::walk_matrix[4*4] = {0,-1,1,-1,
				      -1,0,-1,1,
				      -1,1,0,-1,
				      1,-1,-1,0};

int Crawling_CPG::pace_matrix[4*4] = {0,-1,1,-1,
				      -1,0,-1,1,
				      1,-1,0,-1,
				      -1,1,-1,0};

int Crawling_CPG::bound_matrix[4*4] = {0,1,-1,-1,
				       1,0,-1,-1,
				       -1,-1,0,1,
				       -1,-1,1,0};

int Crawling_CPG::nocoupling_matrix[4*4] = {0,0,0,0,
					    0,0,0,0,
					    0,0,0,0,
					    0,0,0,0};

Crawling_CPG::Crawling_CPG() {
  mDim = 28;
  mTime = new Time;
  *mTime = 0.0;
  mParameters = new Parameter[5];
  mParameters[0] = 1.0; //limit angle mu
  mParameters[1] = 1.0; //omega_swing
  mParameters[2] = 1.0; //omega_stance
  mParameters[3] = 1.0; //alpha
  mParameters[4] = 1.0; //gamma
  
  enable_feedback = 1;
  enable_observ = 1;
  
  select_gait(TROT);

  mu =1;

  for(int i=0;i<4;i++)
    {
      touch_sensors[i] = 0.0;
      current_position[i] = 0.0;
      current_position[2*i] = 0.0;
      discrete_goals[0][i] = 0.0;
      discrete_goals[1][i] = 0.0;
    }
}

Crawling_CPG::~Crawling_CPG() {
	delete mTime;
	delete mParameters;
	
	paramFile.close();
}

void Crawling_CPG::enableFeedback(int en)
{
  enable_feedback = en;
}

void Crawling_CPG::enableObserv(int en)
{
  enable_observ = en;
}

void Crawling_CPG::setDiscreteGoals(double goals[2][4])
{
	for(int i =0;i<2;i++){
		for(int j =0;j<4;j++){
			discrete_goals[i][j] = goals[i][j];
		}
	}
}

void Crawling_CPG::setAmplitude(double amplitude)
{
	mu = amplitude;
}

int Crawling_CPG::select_gait(int gait)
{
  switch(gait)
    {
    case TROT:
      current_gait = trot_matrix;
      return TROT;
    case BOUND:
      current_gait = bound_matrix;
      return BOUND;
    case PACE:
      current_gait = pace_matrix;
      return PACE;
    case WALK:
      current_gait = walk_matrix;
      return WALK;
    case NO_COUPLING:
      current_gait = nocoupling_matrix;
      return NO_COUPLING;
    default:
      return -1;
    }
}

void Crawling_CPG::saveParameters(const char *name)
{
  if(!paramFile.is_open())
    {
      if(name==NULL)
	paramFile.open("CPGparameters.dat",fstream::out);
      else
	paramFile.open(name,fstream::out);
    }
  else
    if(name!=NULL)
      {
	paramFile.close();
	paramFile.open(name,fstream::out);
      }

  for(int i=0;i<2;i++)
    for(int j =0;j<4;j++)
      paramFile << discrete_goals[i][j] << " ";
  
  paramFile << mu ;

  paramFile << endl;
}


void Crawling_CPG::setParameters(Parameter* params)
{
  mParameters[0] = params[0]; //limit angle mu
  mParameters[1] = params[1]; //omega_swing
  mParameters[2] = params[2]; //omega_stance
  mParameters[3] = params[3]; //alpha
  mParameters[4] = params[4]; //gamma
}

void Crawling_CPG::setSensors(short sens[])
{
  for(int i =0;i<4;i++)
    touch_sensors[i] = (double)sens[i];
  //cout << touch_sensors[0] << endl;
}

void Crawling_CPG::setSensors(double sens[])
{
  for(int i =0;i<4;i++)
    touch_sensors[i] = sens[i];
  //cout << touch_sensors[0] << endl;
} 

void Crawling_CPG::setCurrentPosition(double pos[])
{
  for(int i = 0;i<2*4;i++)
    {
      current_position[i] = pos[i];
      //cout << "position " << i << " " << current_position[i] << endl;
    }
}


void Crawling_CPG::odeFunction(double t, const double y[], double dydt[]) {
  
  //mu = (mParameters[0]).getValue();
   double om_sw = (mParameters[1]).getValue();
  double om_st = (mParameters[2]).getValue();
  double alpha = (mParameters[3]).getValue();
  double gamma = (mParameters[4]).getValue();
  double beta = 10.0*alpha;

  double om[4];
  double r[4];
  double b = 100.0;
  
  //discrete system
  //it goes as left:  arm1 12 ,arm2 14, elbow 16 hip1 18
  // then      right: arm1 20 ,arm2 22, elbow 24 hip1 26
  
  int ind = 12;
  for(int i =0;i<8;i++)
    {
      div_t ind2 = div(i,4);
      dydt[ind+2*i]   = y[ind+2*i+1];
      dydt[ind+2*i+1] = gamma*(gamma/4.0 * (discrete_goals[ind2.quot][ind2.rem] - y[ind+2*i]) - y[ind+2*i+1]);
	  //cout<< "y["<< ind+2*i <<"]:" <<y[ind+2*i] <<"\n"; 
    }
  
  /*
  //left arm

	//shoulder 0 (rhythmic) 
  	dydt[12]= y[13];
	dydt[
	13]= gamma*(gamma/4 *(G[0][2]-y[12]) - y[13]);
  */

  
  double y_res[4];
  y_res[0] = y[0] - y[12];
  y_res[1] = y[2] - y[20];
  y_res[2] = y[4] - y[18];
  y_res[3] = y[6] - y[26];

  //update of oscillators without coupling
  for(int i = 0;i<4;i++)
      {
	om[i] = om_st/(exp(-b*y[2*i+1])+1) + y[8+i]/(exp(b*y[2*i+1])+1);
	r[i]  = y_res[i]*y_res[i] + y[2*i+1]*y[2*i+1];
	
	dydt[2*i] = alpha*(mu-r[i])*y_res[i] - om[i]*y[2*i+1];
	dydt[2*i+1] = beta*(mu-r[i])*y[2*i+1] + om[i]*y_res[i];

	 //cout<< "y["<< 2*i <<"]:" <<y[2*i] <<"\n"; 
      }

  //coupling strength
  double strength = 1.0;
  
  //coupling
  for(int i=0;i<4;i++)
    for(int j=0;j<4;j++)
      dydt[2*i+1] += strength*(double)(current_gait[4*i+j])*y[2*j+1];


  
  ////observation
  if(enable_observ)
    {
      double K1 = 20.0;
      double K2 = 50.0;
      double tol = 0.1;
      double tol2 = 0.2;

      for(int i=0;i<4;i++)
	{
	  double err1 = current_position[2*i] - y[2*i];
	  if(err1*err1 < tol*tol)
	    err1 = 0.0;
	  
	  dydt[2*i] += K1*err1;
	  
	  double err2 = -current_position[2*i+1]/om[i] - y[2*i+1];
	  if(err2*err2 < tol2*tol2)
	    err2 = 0.0;

	  dydt[2*i+1] += K2*err2;
	  
	  cout << err1 << " " << err2 << endl;
	}
      
    }
  

  //frequency adaptation
  double epsilon = 0.0;
  for(int i=0;i<4;i++)
    dydt[8+i] = 0.0;
    
}


